@@ -15,6 +15,13 @@ from .generators import Generator1D as _Generator1D
 from .generators import Generator2D as _Generator2D
 from .generators import Generator3D as _Generator3D
 from .conditions import IrregularBoundaryCondition as _IrregularBC
+from .operators import grad
+
+
+def _updatable_contour_plot_available():
+    from packaging.version import parse as vparse
+    from matplotlib import __version__
+    return vparse(__version__) >= vparse('3.3.0')
 
 
 class BaseMonitor(ABC):
@@ -31,6 +38,14 @@ class BaseMonitor(ABC):
     def __init__(self, check_every=None):
         self.check_every = check_every or 100
         self.fig = ...
+        self.using_non_gui_backend = (matplotlib.get_backend() == 'agg')
+
+        if matplotlib.get_backend() == 'module://ipykernel.pylab.backend_inline':
+            warnings.warn(
+                "You seem to be using jupyter notebook with '%matplotlib inline' "
+                "which can lead to monitor plots not updating. "
+                "Consider using '%matplotlib notebook' or '%matplotlib widget' instead.",
+                UserWarning)
 
     @abstractmethod
     def check(self, nets, conditions, history):
@@ -112,11 +127,10 @@ class MonitorSpherical(BaseMonitor):
         """Initializer method
         """
         super(MonitorSpherical, self).__init__(check_every=check_every)
-        self.contour_plot_available = self._matplotlib_version_satisfies()
+        self.contour_plot_available = _updatable_contour_plot_available()
         if not self.contour_plot_available:
             warnings.warn("Warning: contourf plot only available for matplotlib version >= v3.3.0 "
                           "switching to matshow instead")
-        self.using_non_gui_backend = (matplotlib.get_backend() == 'agg')
         self.fig = None
         self.axs = []  # subplots
         self.ax_metrics = None
@@ -150,12 +164,6 @@ class MonitorSpherical(BaseMonitor):
 
         self.n_vars = None
 
-    @staticmethod
-    def _matplotlib_version_satisfies():
-        from packaging.version import parse as vparse
-        from matplotlib import __version__
-        return vparse(__version__) >= vparse('3.3.0')
-
     @staticmethod
     def _longitude_formatter(value, count):
         value = int(round(value / math.pi * 180)) - 180
@@ -340,7 +348,7 @@ class MonitorSpherical(BaseMonitor):
             # see https://github.com/matplotlib/matplotlib/issues/15986
             theta = self.theta_label.reshape(*self.shape)[0, :, 0]
             phi = self.phi_label.reshape(*self.shape)[0, 0, :]
-            cax = ax.contourf(phi, theta, u, cmap='magma')
+            cax = ax.contourf(phi, theta, u, cmap='magma', levels=max(self.shape[-2:]))
             ax.xaxis.set_major_locator(plt.MultipleLocator(math.pi / 6))
             ax.xaxis.set_minor_locator(plt.MultipleLocator(math.pi / 12))
             ax.xaxis.set_major_formatter(plt.FuncFormatter(self._longitude_formatter))
@@ -516,7 +524,6 @@ class Monitor1D(BaseMonitor):
         """Initializer method
         """
         super(Monitor1D, self).__init__(check_every=check_every)
-        self.using_non_gui_backend = (matplotlib.get_backend() == 'agg')
         self.fig = plt.figure(figsize=(30, 8))
         self.ax1 = self.fig.add_subplot(131)
         self.ax2 = self.fig.add_subplot(132)
@@ -584,6 +591,7 @@ class Monitor1D(BaseMonitor):
 
 class Monitor2D(BaseMonitor):
     r"""A monitor for checking the status of the neural network during training.
+    The number and layout of subplots (matplotlib axes) will be finalized after the first ``.check()`` call.
 
     :param xy_min:
         The lower bound of 2 dimensions.
@@ -597,14 +605,57 @@ class Monitor2D(BaseMonitor):
         The frequency of checking the neural network represented by the number of epochs between two checks.
         Defaults to 100.
     :type check_every: int, optional
+    :param valid_generator:
+        The generator used to sample points from the domain when visualizing the solution.
+        The generator is only called once (during instantiating the generator), and its outputs are stored.
+        Defaults to a 32x32 ``Generator2D`` with method 'equally-spaced'.
+    :type valid_generator: neurodiffeq.generators.BaseGenerator
+    :param solution_style:
+
+        - If set to 'heatmap', solution visualization will be a contour heat map of
+          :math:`u` w.r.t. :math:`x` and :math:`y`. Useful when visualizing a 2-D spatial solution.
+        - If set to 'curves', solution visualization will be :math:`u`-:math:`x` curves instead of a 2d heat map.
+          Each curve corresponds to a :math:`t` value. Useful when visualizing 1D spatio-temporal solution.
+          The first coordinate is interpreted as :math:`x` and the second as :math:`t`.
+
+        Defaults to 'heatmap'.
+    :type solution_style: str
+    :param equal_aspect:
+        Whether to set aspect ratio to 1:1 for heatmap. Defaults to True.
+        Ignored if `solutions_style` is 'curves'.
+    :type equal_aspect: bool
+    :param ax_width:
+        Width for each solution visualization. Note that this is different from width for metrics history,
+        which is equal to ``ax_width`` :math:`\times` ``n_cols``.
+    :type ax_width: float
+    :param ax_height: Height for each solution visualization and metrics history plot.
+    :type ax_height: float
+    :param n_col:
+        Number of solution visualizations to plot in each row.
+        Note there is always only 1 plot for metrics history plot per row.
+    :type n_col: int
+    :param levels: Number of levels to plot with contourf (heatmap). Defaults to 20.
+    :type levels: int
     """
 
-    def __init__(self, xy_min, xy_max, check_every=None, valid_generator=None):
+    def __init__(self, xy_min, xy_max, check_every=None, valid_generator=None, solution_style='heatmap',
+                 equal_aspect=True, ax_width=5.0, ax_height=4.0, n_col=2, levels=20):
         """Initializer method
         """
         super(Monitor2D, self).__init__(check_every=check_every)
-        self.using_non_gui_backend = (matplotlib.get_backend() == 'agg')
+        if solution_style not in ['heatmap', 'curves']:
+            raise ValueError(f"Unsupported 'solution_style' = {solution_style}")
+        if not _updatable_contour_plot_available() and solution_style == 'heatmap':
+            warnings.warn("Heatmap-style solution does not work with your matplotlib version. "
+                          "Please upgrade matplotlib to v3.3.0 or higher. "
+                          "Otherwise you may experience buggy behavior.",
+                          UserWarning)
+        self.solution_style = solution_style
         self.fig = None
+        self.ax_width = ax_width
+        self.ax_height = ax_height
+        self.n_col = n_col
+        self.equal_aspect = equal_aspect
         self.axs = []  # subplots
         # self.caxs = []  # colorbars
         self.cbs = []  # color bars
@@ -615,10 +666,10 @@ class Monitor2D(BaseMonitor):
         self.xs_ann, self.ys_ann = xs_ann.reshape(-1, 1), ys_ann.reshape(-1, 1)
         self.xs_plot = self.xs_ann.detach().cpu().numpy().flatten()
         self.ys_plot = self.ys_ann.detach().cpu().numpy().flatten()
+        self.levels = levels
 
     # draw a contour plot of the surface (xs, ys) -> zs
-    @staticmethod
-    def _create_contour(ax, xs, ys, zs, condition):
+    def _create_contour(self, ax, xs, ys, zs, condition):
         triang = tri.Triangulation(xs, ys)
         xs = xs[triang.triangles].mean(axis=1)
         ys = ys[triang.triangles].mean(axis=1)
@@ -628,10 +679,11 @@ class Monitor2D(BaseMonitor):
                 in_domain = condition.in_domain(xs, ys)
                 triang.set_mask(~in_domain)
 
-        contour = ax.tricontourf(triang, zs, cmap='coolwarm')
+        contour = ax.tricontourf(triang, zs, cmap='coolwarm', levels=self.levels)
         ax.set_xlabel('x')
         ax.set_ylabel('y')
-        ax.set_aspect('equal', adjustable='box')
+        if self.equal_aspect:
+            ax.set_aspect('equal', adjustable='box')
         return contour
 
     def check(self, nets, conditions, history):
@@ -658,29 +710,38 @@ class Monitor2D(BaseMonitor):
             # size of the figure, number of the subplots, etc.
 
             # one for each dependent variable, plus one for training and validation loss, plus one for metrics
-            n_axs = len(conditions) + 2
-            n_row, n_col = (n_axs + 1) // 2, 2
-            self.fig = plt.figure(figsize=(20, 8 * n_row))
-            for i in range(n_axs):
+            n_func = len(conditions)
+            n_col = self.n_col
+            n_row_sols = math.ceil(n_func / n_col)
+            n_row = n_row_sols + 2
+            self.fig = plt.figure(figsize=(self.ax_width * n_col, self.ax_height * n_row))
+            self.fig.tight_layout()
+            # axes and color bars for solutions (aka dependent variables)
+            for i in range(n_func):
                 self.axs.append(self.fig.add_subplot(n_row, n_col, i + 1))
-            for i in range(n_axs - 2):
                 self.cbs.append(None)
+            # axes for history plot of loss and other metrics, these plots should take the whole row
+            self.axs.append(self.fig.add_subplot(n_row, 1, n_row_sols + 1))
+            self.axs.append(self.fig.add_subplot(n_row, 1, n_row_sols + 2))
 
         us = [
             con.enforce(net, self.xs_ann, self.ys_ann)
             for con, net in zip(conditions, nets)
         ]
 
-        for i, ax_u_con in enumerate(zip(self.axs[:-2], us, conditions)):
-            ax, u, con = ax_u_con
+        for i, (ax, u, con) in enumerate(zip(self.axs[:-2], us, conditions)):
             ax.clear()
             u = u.detach().cpu().numpy().flatten()
-            cs = self._create_contour(ax, self.xs_plot, self.ys_plot, u, con)
-            if self.cbs[i] is None:
+            if self.solution_style == 'heatmap':
+                cs = self._create_contour(ax, self.xs_plot, self.ys_plot, u, con)
+                if self.cbs[i] is not None:
+                    self.cbs[i].remove()
                 self.cbs[i] = self.fig.colorbar(cs, format='%.0e', ax=ax)
-            else:
-                self.cbs[i].mappable.set_clim(vmin=u.min(), vmax=u.max())
-            ax.set_title(f'u[{i}](x, y)')
+                ax.set_title(f'u[{i}](x, y)')
+            elif self.solution_style == 'curves':
+                df = pd.DataFrame(dict(u=u, x=self.xs_plot, t=self.ys_plot))
+                sns.lineplot(x='x', y='u', data=df, hue='t', ax=ax)
+                ax.set_title(f'u[{i}](x) across different t')
 
         self.axs[-2].clear()
         self.axs[-2].plot(history['train_loss'], label='training loss')
@@ -707,3 +768,136 @@ class Monitor2D(BaseMonitor):
         self.fig.canvas.draw()
         if not self.using_non_gui_backend:
             plt.pause(0.05)
+
+
+class MetricsMonitor(BaseMonitor):
+    r"""A monitor for visualizing the loss and other metrics.
+    This monitor does not visualize the solution.
+
+    :param check_every:
+        The frequency of checking the neural network represented by the number of epochs between two checks.
+        Defaults to 100.
+    :type check_every: int, optional
+    """
+
+    def __init__(self, check_every=None):
+        super().__init__(check_every=check_every)
+        self.fig = plt.figure(figsize=(12, 6), dpi=125)
+        self.ax1, self.ax2 = self.fig.subplots(1, 2)
+
+    def check(self, nets, conditions, history):
+        self.ax1.clear()
+        self.ax1.plot(history['train_loss'], label='training loss')
+        self.ax1.plot(history['valid_loss'], label='validation loss')
+        self.ax1.set_title('loss during training')
+        self.ax1.set_ylabel('loss')
+        self.ax1.set_xlabel('epochs')
+        self.ax1.set_yscale('log')
+        self.ax1.legend()
+
+        self.ax2.clear()
+        for metric_name, metric_values in history.items():
+            if metric_name == 'train_loss' or metric_name == 'valid_loss':
+                continue
+            self.ax2.plot(metric_values, label=metric_name)
+        self.ax2.set_title('metrics during training')
+        self.ax2.set_ylabel('metrics')
+        self.ax2.set_xlabel('epochs')
+        self.ax2.set_yscale('log')
+        # if there're not custom metrics, then there won't be any labels in this axis
+        if len(history) > 2:
+            self.ax2.legend()
+
+        self.fig.canvas.draw()
+        if not self.using_non_gui_backend:
+            plt.pause(0.05)
+
+
+class StreamPlotMonitor2D(BaseMonitor):
+    def __init__(self, xy_min, xy_max, pairs, nx=32, ny=32, check_every=None, mask_fn=None,
+                 ax_width=13.0, ax_height=10.0, n_col=2, stream_kwargs=None, equal_aspect=True, field_names=None):
+        super(StreamPlotMonitor2D, self).__init__(check_every=check_every)
+        self.pairs = pairs
+        self.field_names = field_names or [f'Field[{i}]' for i, _ in enumerate(pairs)]
+        if len(self.field_names) != len(self.pairs):
+            raise ValueError(f"Length of field_names ({len(self.field_names)}) != Length of pairs({len(self.pairs)})")
+        n_row = int(np.ceil(len(self.pairs) / n_col))
+        self.nx, self.ny = nx, ny
+        self.fig = plt.figure(figsize=(n_col * ax_width, n_row * ax_height))
+        self.axes = self.fig.subplots(n_row, n_col).reshape(-1)
+        self.cbs = [None] * len(pairs)  # colorbars
+        _xs_ann, _ys_ann = torch.meshgrid([
+            torch.linspace(xy_min[0], xy_max[0], nx, requires_grad=True),
+            torch.linspace(xy_min[1], xy_max[1], ny, requires_grad=True),
+        ])
+        self.xs_ann, self.ys_ann = _xs_ann.reshape(-1, 1), _ys_ann.reshape(-1, 1)
+        self.xs_plot = _xs_ann.detach().cpu().numpy()
+        self.ys_plot = _ys_ann.detach().cpu().numpy()
+        self.xlim = xy_min[0], xy_max[0]
+        self.ylim = xy_min[1], xy_max[1]
+
+        if mask_fn:
+            self.mask = mask_fn(self.xs_plot, self.ys_plot)
+            # TODO use antialiasing
+            _pcolor_x, _pcolor_y = np.meshgrid(
+                np.linspace(xy_min[0], xy_max[0], nx * 8),
+                np.linspace(xy_min[1], xy_max[1], ny * 8),
+            )
+            _pcolor_mask = mask_fn(_pcolor_x, _pcolor_y)
+            self._pcolor_args = (_pcolor_x, _pcolor_y, ~_pcolor_mask)
+        else:
+            self.mask = None
+            self._pcolor_args = ()
+        self.stream_kwargs = dict(
+            density=(self.nx / 30, self.ny / 30),
+        )
+        self.stream_kwargs.update(stream_kwargs or {})
+        self.equal_aspect = equal_aspect
+
+    def _plot_streamlines(self, ax, us, vs, norms, cb_idx, is_grad=False):
+        ax.clear()
+        if self.mask is not None:
+            # FIXME if mask covers all points in the meshgrid, the following ValueError will be raised
+            # "Need at least one array to concatenate"
+            us[~self.mask] = np.nan
+            vs[~self.mask] = np.nan
+            ax.pcolor(*self._pcolor_args, shading='auto', cmap='Purples')
+        kwargs = dict(color=norms.transpose())
+        kwargs.update(self.stream_kwargs)
+        stream = ax.streamplot(self.xs_plot[:, 0], self.ys_plot[0, :], us.transpose(), vs.transpose(), **kwargs)
+        if self.cbs[cb_idx] is not None:
+            self.cbs[cb_idx].remove()
+        self.cbs[cb_idx] = self.fig.colorbar(stream.lines, ax=ax)
+        if self.equal_aspect:
+            ax.set_aspect('equal', adjustable='box')
+        ax.set_xlim(*self.xlim)
+        ax.set_ylim(*self.ylim)
+        if is_grad:
+            ax.set_title(f'Gradient of {self.field_names[cb_idx]}')
+        else:
+            ax.set_title(f'Stream Plot of {self.field_names[cb_idx]}')
+
+    def check(self, nets, conditions, history):
+        for idx, pair in enumerate(self.pairs):
+            if isinstance(pair, int):
+                p = conditions[pair].enforce(nets[pair], self.xs_ann, self.ys_ann)
+                us, vs = grad(p, self.xs_ann, self.ys_ann)
+                us = us.reshape(self.nx, self.ny)
+                vs = vs.reshape(self.nx, self.ny)
+                is_grad = True
+            else:
+                ui, vi = pair
+                us = conditions[ui].enforce(nets[ui], self.xs_ann, self.ys_ann).reshape(self.nx, self.ny)
+                vs = conditions[vi].enforce(nets[vi], self.xs_ann, self.ys_ann).reshape(self.nx, self.ny)
+                is_grad = False
+
+            norms = torch.sqrt(us ** 2 + vs ** 2)
+
+            self._plot_streamlines(
+                ax=self.axes[idx],
+                us=us.detach().cpu().numpy(),
+                vs=vs.detach().cpu().numpy(),
+                norms=norms.detach().cpu().numpy(),
+                cb_idx=idx,
+                is_grad=is_grad,
+            )
