@@ -281,8 +281,8 @@ def solve_system(
 
     if additional_loss_term:
         class CustomSolver1D(Solver1D):
-            def additional_loss(self, funcs, key):
-                return additional_loss_term(*funcs, *self._batch_examples[key])
+            def additional_loss(self, residual, funcs, coords):
+                return additional_loss_term(*funcs, *coords)
     else:
         class CustomSolver1D(Solver1D):
             pass
