@@ -7,15 +7,17 @@ from abc import ABC, abstractmethod
 from itertools import chain
 from copy import deepcopy
 from torch.optim import Adam
-from neurodiffeq.networks import FCNN
-from neurodiffeq._version_utils import deprecated_alias
-from neurodiffeq.generators import GeneratorSpherical
-from neurodiffeq.generators import SamplerGenerator
-from neurodiffeq.generators import Generator1D
-from neurodiffeq.generators import Generator2D
-from neurodiffeq.generators import GeneratorND
-from neurodiffeq.function_basis import RealSphericalHarmonics
-from neurodiffeq.neurodiffeq import safe_diff as diff
+from .solvers_utils import PretrainedSolver
+from .networks import FCNN
+from ._version_utils import deprecated_alias
+from .generators import GeneratorSpherical
+from .generators import SamplerGenerator
+from .generators import Generator1D
+from .generators import Generator2D
+from .generators import GeneratorND
+from .function_basis import RealSphericalHarmonics
+from .conditions import BaseCondition
+from .neurodiffeq import safe_diff as diff
 from .losses import _losses
 
 
@@ -23,7 +25,7 @@ def _requires_closure(optimizer):
     return inspect.signature(optimizer.step).parameters.get('closure').default == inspect._empty
 
 
-class BaseSolver(ABC):
+class BaseSolver(ABC, PretrainedSolver):
     r"""A class for solving ODE/PDE systems.
 
     :param diff_eqs:
@@ -48,15 +50,26 @@ class BaseSolver(ABC):
     :param analytic_solutions:
         **[DEPRECATED]** Pass ``metrics`` instead.
         The analytical solutions to be compared with neural net solutions.
-        It maps a tuple of three coordinates to a tuple of function values.
+        It maps a tuple of coordinates to a tuple of function values.
         The output shape should match that of networks.
     :type analytic_solutions: callable, optional
     :param optimizer:
         The optimizer to be used for training.
     :type optimizer: `torch.nn.optim.Optimizer`, optional
     :param criterion:
-        A function that maps a PDE residual vector (torch tensor with shape (-1, 1)) to a scalar loss.
-    :type criterion: callable, optional
+        The loss function used for training.
+
+        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
+        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
+        - If any other callable, it must map
+          A) a residual tensor (shape `(n_points, n_equations)`),
+          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
+          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
+          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
+          so that backpropagation can be performed.
+
+    :type criterion:
+        str or `torch.nn.moduesl.loss._Loss` or callable
     :param n_batches_train:
         Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
         Defaults to 1.
@@ -159,15 +172,7 @@ class BaseSolver(ABC):
         self.metrics_history.update({'valid__' + name: [] for name in self.metrics_fn})
 
         self.optimizer = optimizer if optimizer else Adam(chain.from_iterable(n.parameters() for n in self.nets))
-
-        if criterion is None:
-            self.criterion = lambda r, f, x: (r ** 2).mean()
-        elif isinstance(criterion, nn.modules.loss._Loss):
-            self.criterion = lambda r, f, x: criterion(r, torch.zeros_like(r))
-        elif isinstance(criterion, str):
-            self.criterion = _losses[criterion.lower()]
-        else:
-            self.criterion = criterion
+        self._set_criterion(criterion)
 
         def make_pair_dict(train=None, valid=None):
             return {'train': train, 'valid': valid}
@@ -194,6 +199,18 @@ class BaseSolver(ABC):
         # the _phase variable is registered for callback functions to access
         self._phase = None
 
+    def _set_criterion(self, criterion):
+        if criterion is None:
+            self.criterion = lambda r, f, x: (r ** 2).mean()
+        elif isinstance(criterion, nn.modules.loss._Loss):
+            self.criterion = lambda r, f, x: criterion(r, torch.zeros_like(r))
+        elif isinstance(criterion, str):
+            self.criterion = _losses[criterion.lower()]
+        elif callable(criterion):
+            self.criterion = criterion
+        else:
+            raise TypeError(f"Unknown type of criterion {type(criterion)}")
+
     @property
     def global_epoch(self):
         r"""Global epoch count, always equal to the length of train loss history.
@@ -681,8 +698,19 @@ class SolverSpherical(BaseSolver):
         Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
     :type optimizer: ``torch.nn.optim.Optimizer``, optional
     :param criterion:
-        Function that maps a PDE residual tensor (of shape (-1, 1)) to a scalar loss.
-    :type criterion: callable, optional
+        The loss function used for training.
+
+        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
+        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
+        - If any other callable, it must map
+          A) a residual tensor (shape `(n_points, n_equations)`),
+          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
+          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
+          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
+          so that backpropagation can be performed.
+
+    :type criterion:
+        str or `torch.nn.moduesl.loss._Loss` or callable
     :param n_batches_train:
         Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
         Defaults to 1.
@@ -691,6 +719,14 @@ class SolverSpherical(BaseSolver):
         Number of batches to validate in every epoch, where batch-size equals ``valid_generator.size``.
         Defaults to 4.
     :type n_batches_valid: int, optional
+    :param metrics:
+        Additional metrics to be logged (besides loss). ``metrics`` should be a dict where
+
+        - Keys are metric names (e.g. 'analytic_mse');
+        - Values are functions (callables) that computes the metric value.
+          These functions must accept the same input as the differential equation ``diff_eq``.
+
+    :type metrics: dict, optional
     :param enforcer:
         A function of signature
         ``enforcer(net: nn.Module, cond: neurodiffeq.conditions.BaseCondition,
@@ -715,7 +751,7 @@ class SolverSpherical(BaseSolver):
 
     def __init__(self, pde_system, conditions, r_min=None, r_max=None,
                  nets=None, train_generator=None, valid_generator=None, analytic_solutions=None,
-                 optimizer=None, criterion=None, n_batches_train=1, n_batches_valid=4, enforcer=None,
+                 optimizer=None, criterion=None, n_batches_train=1, n_batches_valid=4, metrics=None, enforcer=None,
                  n_output_units=1,
                  # deprecated arguments are listed below
                  shuffle=None, batch_size=None):
@@ -746,6 +782,7 @@ class SolverSpherical(BaseSolver):
             criterion=criterion,
             n_batches_train=n_batches_train,
             n_batches_valid=n_batches_valid,
+            metrics=metrics,
             n_input_units=3,
             n_output_units=n_output_units,
             shuffle=shuffle,
@@ -768,7 +805,11 @@ class SolverSpherical(BaseSolver):
         if self.enforcer:
             return self.enforcer(net, cond, coordinates)
 
-        n_params = len(signature(cond.enforce).parameters)
+        # Base .enforce takes a variable length *arg; n_params should be deduced from .parameterize
+        if cond.__class__.enforce == BaseCondition.enforce:
+            n_params = len(signature(cond.parameterize).parameters)
+        else:
+            n_params = len(signature(cond.enforce).parameters)
         coordinates = coordinates[:n_params - 1]
         return cond.enforce(net, *coordinates)
 
@@ -916,8 +957,19 @@ class Solver1D(BaseSolver):
         Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
     :type optimizer: ``torch.nn.optim.Optimizer``, optional
     :param criterion:
-        Function that maps a ODE residual tensor (of shape (-1, 1)) to a scalar loss.
-    :type criterion: callable, optional
+        The loss function used for training.
+
+        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
+        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
+        - If any other callable, it must map
+          A) a residual tensor (shape `(n_points, n_equations)`),
+          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
+          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
+          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
+          so that backpropagation can be performed.
+
+    :type criterion:
+        str or `torch.nn.moduesl.loss._Loss` or callable
     :param n_batches_train:
         Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
         Defaults to 1.
@@ -1053,13 +1105,15 @@ class BundleSolver1D(BaseSolver):
     :type t_max: float, optional
     :param theta_min:
         Lower bound of input (parameters and/or conditions). If conditions are included in the bundle,
-        they should be the first values in the tuple.
+        the order should match the one inferred by the values of the ``bundle_conditions`` input
+        in the ``neurodiffeq.conditions.BundleIVP``.
         Defaults to None.
         Ignored if ``train_generator`` and ``valid_generator`` are both set.
     :type theta_min: float or tuple, optional
     :param theta_max:
         Upper bound of input (parameters and/or conditions). If conditions are included in the bundle,
-        they should be the first values in the tuple.
+        the order should match the one inferred by the values of the ``bundle_conditions`` input
+        in the ``neurodiffeq.conditions.BundleIVP``.
         Defaults to None.
         Ignored if ``train_generator`` and ``valid_generator`` are both set.
     :type theta_max: float or tuple, optional
@@ -1087,8 +1141,19 @@ class BundleSolver1D(BaseSolver):
         Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
     :type optimizer: ``torch.nn.optim.Optimizer``, optional
     :param criterion:
-        Function that maps a ODE residual tensor (of shape (-1, 1)) to a scalar loss.
-    :type criterion: callable, optional
+        The loss function used for training.
+
+        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
+        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
+        - If any other callable, it must map
+          A) a residual tensor (shape `(n_points, n_equations)`),
+          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
+          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
+          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
+          so that backpropagation can be performed.
+
+    :type criterion:
+        str or `torch.nn.moduesl.loss._Loss` or callable
     :param n_batches_train:
         Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
         Defaults to 1.
@@ -1134,35 +1199,51 @@ class BundleSolver1D(BaseSolver):
                                  f"got t_min={t_min}, t_max={t_max}, "
                                  f"train_generator={train_generator}, valid_generator={valid_generator}")
 
-        if isinstance(theta_min, float) or isinstance(theta_min, int):
-            r_min = (t_min,) + (theta_min,)
+        if isinstance(theta_min, (float, int)):
+            theta_min = (theta_min,)
 
-        if isinstance(theta_max, float) or isinstance(theta_max, int):
-            r_max = (t_max,) + (theta_max,)
+        if isinstance(theta_max, (float, int)):
+            theta_max = (theta_max,)
 
         if theta_min is None and theta_max is None:
             r_min = (t_min,)
             r_max = (t_max,)
-
-        if isinstance(theta_min, tuple) and isinstance(theta_max, tuple):
+        else:
             r_min = (t_min,) + theta_min
             r_max = (t_max,) + theta_max
 
         n_input_units = len(r_min)
 
-        methods = ['equally-spaced' for i in range(n_input_units)]
-
-        grid = tuple(32 for j in range(n_input_units))
-
         if train_generator is None:
-            train_generator = GeneratorND(grid, r_min=r_min, r_max=r_max, methods=methods, noisy=True)
+            train_generator = Generator1D(32, t_min=t_min, t_max=t_max, method='equally-spaced-noisy')
+            for i in range(n_input_units - 1):
+                train_generator ^= Generator1D(32, t_min=r_min[i + 1], t_max=r_max[i + 1],
+                                               method='equally-spaced-noisy')
         if valid_generator is None:
-            valid_generator = GeneratorND(grid, r_min=r_min, r_max=r_max, methods=methods, noisy=False)
+            valid_generator = Generator1D(32, t_min=t_min, t_max=t_max, method='equally-spaced')
+            for i in range(n_input_units - 1):
+                valid_generator ^= Generator1D(32, t_min=r_min[i + 1], t_max=r_max[i + 1], method='equally-spaced')
+
+        non_var = []
+        for c in conditions:
+            try:
+                bc = tuple(c.bundle_conditions.values())
+            except Exception:
+                bc = ()
+            for index in bc:
+                non_var.append(index)
+        non_var = list(set(non_var))
+
+        def non_var_filter(*variables):
+            variables = list(variables)
+            for i, n in enumerate(non_var):
+                del variables[len(conditions) + 1 + n - i]
+            return ode_system(*variables)
 
         self.r_min, self.r_max = r_min, r_max
 
         super(BundleSolver1D, self).__init__(
-            diff_eqs=ode_system,
+            diff_eqs=non_var_filter,
             conditions=conditions,
             nets=nets,
             train_generator=train_generator,
@@ -1271,8 +1352,19 @@ class Solver2D(BaseSolver):
         Defaults to a ``torch.optim.Adam`` instance that trains on all parameters of ``nets``.
     :type optimizer: ``torch.nn.optim.Optimizer``, optional
     :param criterion:
-        Function that maps a PDE residual tensor (of shape (-1, 1)) to a scalar loss.
-    :type criterion: callable, optional
+        The loss function used for training.
+
+        - If a str, must be present in the keys of `neurodiffeq.losses._losses`.
+        - If a `torch.nn.modules.loss._Loss` instance, just pass the instance.
+        - If any other callable, it must map
+          A) a residual tensor (shape `(n_points, n_equations)`),
+          B) a function values tuple (length `n_funcs`, each element a tensor of shape `(n_points, 1)`), and
+          C) a coordinate values tuple (length `n_coords`, each element a tensor of shape `(n_coords, 1)`
+          to a tensor of empty shape (i.e. a scalar). The returned tensor must be connected to the computational graph,
+          so that backpropagation can be performed.
+
+    :type criterion:
+        str or `torch.nn.moduesl.loss._Loss` or callable
     :param n_batches_train:
         Number of batches to train in every epoch, where batch-size equals ``train_generator.size``.
         Defaults to 1.
